#include <libc.h>
#include <bsd/errno.h>
#include <stdio.h>
#include <ctype.h>

#include "clio.h"

#undef PARSE_DEBUG
#define TRACE_PAIRS
#define MAXTOK (256)

	SVGAIndexValuePair      pairs[CLIO_MAXPAIRS];
	int                     currentPair;
	CirrusLogicGD754XMode   mode;
	char                    name[MAXTOK];
	
	typedef enum {
		identifier,number,hexnumber,string,special,eof,error
	}			TokenKindT;

#ifdef PARSE_DEBUG
	const char* tokenKinds[]={
			"identifier","number","hexnumber","string","special","eof","error"
		};
#endif

	short                   parseAbort;
	const char*             filename;
	const char*				source;
	const char*				current;
	int                     line;
	char                    token[MAXTOK];
	TokenKindT              tokenKind;

/*
	registers   ::= name general sequencer controler graphic attribute end .
	name        ::= 'NAME' ':' string .
	general     ::= 'GENERAL' ':' pairs .
	sequencer   ::= 'SEQUENCER' ':' pairs .
	controler   ::= 'CONTROLER' ':' pairs .
	graphic     ::= 'GRAPHIC'   ':' pairs .
	attribute   ::= 'ATTRIBUTE' ':' pairs .
	end         ::= 'END' '.' .
	string      ::= '"' {character} '"' .
	pairs       ::= value ',' value ',' pairs .
	pairs       ::= .
	value       ::= '0x'{hexadecimal} | {decimal} .
	hexadecimal ::= 0|1||2|3|4|5|6|7|8|9|A|a|B|b|C|c|D|d|E|e|F|f
*/	

	short parseInitialize(FILE* regfile)
	{
			long	size;

		parseAbort=1;
		if(fseek(regfile,0,SEEK_END)!=0){
			fprintf(stderr,"Cannot seek to the end-of-file.\n");
			return(0);
		}
		size=ftell(regfile);
		if(fseek(regfile,0,SEEK_SET)!=0){
			fprintf(stderr,"Cannot seek to the begin-of-file.\n");
			return(0);
		}
		source=(const char*)malloc(size+1);
		if(fread((char*)source,sizeof(char),size,regfile)!=size){
			fprintf(stderr,"Cannot read to the file.\n");
			return(0);
		}
		((char*)source)[size]='\0';
		current=source;
		line=1;
		parseAbort=0;
		currentPair=0;
		mode.name=&(name[0]);
		return(1);
	}//parseInitialize;


	void parseTerminate(void)
	{
		free((char*)source);
		source=0;
		current=0;
	}//parseTerminate;
	

	void parseError(const char* errorMessage)
	{
		fprintf(stderr,"%s:%d: %s\n",filename,line,errorMessage);
		tokenKind=error;
		parseAbort=1;
	}//parseError;
	

	void parseToken(void)
	{
			int begline;
			int i;
			char message[256];
	begin:	
		i=0;
		while((*current!='\0')&&(isspace(*current))){
			if(*current=='\n'){
				line++;
			}
			current++;
		}
		if(isalpha(*current)){
			while((i<MAXTOK)&&(isalpha(*current)||isdigit(*current))){
				token[i++]=*current++;
			}
			tokenKind=identifier;
		}else if(isdigit(*current)){
			if((*current=='0')&&(current[1]=='x')){
				token[0]=current[0];
				token[1]=current[1];
				current+=2;
				i=2;
				while((i<MAXTOK)&&(isxdigit(*current))){
					token[i++]=*current++;
				}
				tokenKind=hexnumber;
			}else{
				while((i<MAXTOK)&&(isdigit(*current))){
					token[i++]=*current++;
				}
				tokenKind=number;
			}
		}else{
			switch(*current){
			case '\0':
				tokenKind=eof;
				break;
			case ':':
			case ',':
			case '.':
				token[i++]=*current++;
				tokenKind=special;
				break;
			case '"':
				begline=line;
				current++;
				while((i<MAXTOK)&&(*current!='\0')
				&&(*current!='"')&&(*current!='\n')){
					token[i++]=*current++;
				}
				if(*current=='"'){
					current++;
					tokenKind=string;
				}else{
					if(i<MAXTOK){
						tokenKind=error;
						parseError("found an unbounded string");
					}
					if(*current=='\n'){
						current++;
						line++;
					}
				}
				break;
			case '/':
				if(current[1]=='/'){
					while((*current!='\0')&&(*current!='\n')){
						current++;
					}
					if(*current=='\n'){
						current++;
						line++;
					}
					goto begin;
				}//else fall thru!
			default:
				sprintf(message,"Invalid character: '%c' (%d)",
								*current,*current);
				parseError(message);
				break;
			}
		}
		if(i<MAXTOK){
			token[i]='\0';
		}else{
			parseError("found a token too large");
			token[MAXTOK-1]='\0';
		}
#ifdef PARSE_DEBUG
		printf("parseToken: found %12s '%s'\n",tokenKinds[tokenKind],token);
#endif
	}//parseToken;
	
	
	
	void parseName(void)
		// name        ::= 'NAME' ':' string .
	{
		parseToken(); // skip KEYWORD
		if((tokenKind!=special)||(strcmp(token,":")!=0)){
			parseError("':' expected");
		}else{
			parseToken(); // skip ':'
			if(tokenKind!=string){
				parseError("a string was expected");
			}else{
				strcpy(name,token);
				parseToken();
			}
		}
	}//parseName;
	
	
	int parseValue(void)
	{
			int value=0;
			
		switch(tokenKind){
		case number:
			if(sscanf(token,"%d",&value)!=1){
				parseError("found an invalid decimal number");
			}else{
				parseToken();
			}
			break;
		case hexnumber:
			if(sscanf(token,"%x",&value)!=1){
				parseError("found an invalid decimal number");
			}else{
				parseToken();
			}
			break;
		default:
			parseError("a number was expected");
			break;
		}	
		return(value);
	}//parseValue;
	
	
	void parsePairs(void)
		// pairs       ::= value ',' value ',' pairs .
		// pairs       ::= .
	{
			int	value;
#ifdef PARSE_DEBUG
		printf("parsePairs\n");
#endif
		do{
			pairs[currentPair].index=parseValue();
			if((tokenKind!=special)||(strcmp(token,",")!=0)){
				parseError("a comma is missing");
			}else{
				parseToken(); // skip ','.
				pairs[currentPair].value=parseValue();
				pairs[currentPair].filler=0xa5;
				if((tokenKind!=special)||(strcmp(token,",")!=0)){
					parseError("a comma is missing");
				}else{
#if defined(PARSE_DEBUG)||defined(TRACE_PAIRS)
					printf("parsePairs:      %04x  %02x\n",
							pairs[currentPair].index,pairs[currentPair].value);
#endif
					currentPair++;
					if(currentPair>=CLIO_MAXPAIRS){
						char message[256];
						sprintf(message,"too many pairs (current max=%d)",
									CLIO_MAXPAIRS);
						parseError(message);
					}else{
						parseToken(); // skip ','.
					}
				}
			}
		}while((tokenKind==number)||(tokenKind==hexnumber));
	}//parsePairs;
	

	void parseKeyword(SVGAIndexValuePair** registers)
		// keyword     ::= KEYWORD ':' pairs .
	{
#ifdef PARSE_DEBUG
		printf("parseKeyword\n");
#endif
#if defined(TRACE_PAIRS)
		printf("parsePairs: %s registers\n",token);
#endif
		parseToken(); // skip KEYWORD
		if((tokenKind!=special)||(strcmp(token,":")!=0)){
			parseError("':' expected");
		}else{
			parseToken(); // skip ':'
			(*registers)=&(pairs[currentPair]);
			parsePairs();
			pairs[currentPair].index=-1;
			pairs[currentPair].value=0;
			pairs[currentPair].filler=0xa5;
			currentPair++;
		}
	}//parseKeyword;
	

	CirrusLogicGD754XMode* parseRegisters(FILE* regfile)
	// registers ::= name general sequencer controler graphic attribute end.
	{
		if(parseInitialize(regfile)){
			parseToken();
			if((tokenKind!=identifier)||(strcmp(token,"NAME")!=0)){
				parseError("NAME expected");
			}else{
				parseName();
				if((tokenKind!=identifier)||(strcmp(token,"GENERAL")!=0)){
					parseError("GENERAL expected");
				}else{
					parseKeyword(&(mode.generalRegisters));
					if((tokenKind!=identifier)
					||(strcmp(token,"SEQUENCER")!=0)){
						parseError("SEQUENCER expected");
					}else{
						parseKeyword(&(mode.sequencerRegisters));
						if((tokenKind!=identifier)
						||(strcmp(token,"CRT")!=0)){
							parseError("CRT expected");
						}else{
							parseKeyword(&(mode.crtControllerRegisters));
							if((tokenKind!=identifier)
							||(strcmp(token,"GRAPHIC")!=0)){
								parseError("GRAPHIC expected");
							}else{
								parseKeyword(
										&(mode.graphicsControllerRegisters));
								if((tokenKind!=identifier)
								||(strcmp(token,"ATTRIBUTE")!=0)){
									parseError("ATTRIBUTE expected");
								}else{
									parseKeyword(
										&(mode.attributeControllerRegisters));
									if((tokenKind!=identifier)
									||(strcmp(token,"END")!=0)){
										parseError("END expected");
									}else{
										parseToken();
										if((tokenKind!=special)
										||(strcmp(token,".")!=0)){
											parseError("'.' expected");
											parseToken();
											if(tokenKind!=eof){
												parseError("EOF expected");
											}
										}
									}
								}
							}
						}
					}
				}
			}
			parseTerminate();
		}
		if(parseAbort){
			return(NULL);
		}else{
			return(&mode);
		}
	}//parseRegisters;
	


int main(int argc,const char** argv)
{
		int error=0;
		int cirrus;
		int level;
	
	if(argc<3){
		error=7;
	}else{
		cirrus=open("/dev/cirrus",O_RDWR,0);
		if(cirrus<0){
			perror("openning cirrus device:");
			return(9);
		}
		if(strcmp(argv[1],"setbrightness")==0){
			if(argc!=3){
				error=8;
			}else if(isdigit(argv[2][0])){
				level=atoi(argv[2]);
				if(ioctl(cirrus,CLIOCSETBRIGHT,(char*)&level)!=0){
					perror("cirrus ioctl returned:");
					error=10;
				}
			}else{
				error=2;
			}
		}else if(strcmp(argv[1],"outb")==0){
			if(argc!=4){
				error=8;
			}else{
				int             temp;
				clio_inoutb_t   param;
				if(sscanf(argv[2],"%x",&temp)!=1){
					error=2;
				}
				param.port=temp;
				if(sscanf(argv[3],"%x",&temp)!=1){
					error=3;
				}
				param.byte=temp;
				if(error==0){
					if(ioctl(cirrus,CLIOCOUTB,(char*)&param)!=0){
						perror("cirrus ioctl returned:");
						error=10;
					}
				}
			}
		}else if(strcmp(argv[1],"inb")==0){
			if(argc!=3){
				error=2;
			}else{
				int             temp;
				clio_inoutb_t   param;
				if(sscanf(argv[2],"%x",&temp)!=1){
					error=2;
				}
				param.port=temp;
				if(error==0){
					if(ioctl(cirrus,CLIOCINB,(char*)&param)!=0){
						perror("cirrus ioctl returned:");
						error=10;
					}else{
						printf("%x\n",param.byte);
					}
				}
			}
		}else if(strcmp(argv[1],"setregisters")==0){
			if(argc!=3){
				error=2;
			}else{
				FILE* registers;
				filename=argv[2];
				registers=fopen(filename,"r");
				if(registers==NULL){
					fprintf(stderr,"Cannot open the file '%s'\n",filename);
					error=6;
				}else{
					CirrusLogicGD754XMode* mode=parseRegisters(registers);
					fclose(registers);
					if(mode!=NULL){
						if(ioctl(cirrus,CLIOCSETREG,(char*)mode)!=0){
							perror("cirrus ioctl returned:");
							error=10;
						}else{
							printf("Registers downloaded.\n");
						}
					}// won't free the mode.
				}
			}
		}else{
			error=1;
		}
		close(cirrus);
	}
	if((0<error)&&(error<10)){
		fprintf(stderr,"usage:\n"
				"\t%s setbrightness _level_\n"
				"\t%s outb _port_ _byte_\n"
				"\t%s inb _port_\n"
				"\t%s setregisters _regfile_\n"
				" (_port_ and _byte_ in hexadecimal, _level_ in decimal).\n",
					argv[0],argv[0],argv[0],argv[0]);
	}
	return(error);
}/*main*/